@@ -31,7 +31,7 @@
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/commands/vacuumlazy.c,v 1.50.4.5 2007/09/16 02:38:14 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/commands/vacuumlazy.c,v 1.61.2.5 2007/09/16 02:38:02 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -44,10 +44,12 @@
 #include "access/xlog.h"
 #include "commands/vacuum.h"
 #include "miscadmin.h"
+#include "pgstat.h"
 #include "storage/freespace.h"
-#include "storage/sinval.h"
 #include "storage/smgr.h"
 #include "utils/lsyscache.h"
+#include "utils/memutils.h"
+#include "utils/pg_rusage.h"
 
 
 /*
@@ -60,16 +62,13 @@
 #define REL_TRUNCATE_MINIMUM	1000
 #define REL_TRUNCATE_FRACTION	16
 
-/* MAX_TUPLES_PER_PAGE can be a conservative upper limit */
-#define MAX_TUPLES_PER_PAGE		((int) (BLCKSZ / sizeof(HeapTupleHeaderData)))
-
 
 typedef struct LVRelStats
 {
 	/* Overall statistics about rel */
 	BlockNumber rel_pages;
 	double		rel_tuples;
-	BlockNumber	pages_removed;
+	BlockNumber pages_removed;
 	double		tuples_deleted;
 	BlockNumber nonempty_pages; /* actually, last nonempty page + 1 */
 	Size		threshold;		/* minimum interesting free space */
@@ -99,9 +98,9 @@ static void lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 static void lazy_vacuum_heap(Relation onerel, LVRelStats *vacrelstats);
 static void lazy_scan_index(Relation indrel, LVRelStats *vacrelstats);
 static void lazy_vacuum_index(Relation indrel,
-							  double *index_tups_vacuumed,
-							  BlockNumber *index_pages_removed,
-							  LVRelStats *vacrelstats);
+				  double *index_tups_vacuumed,
+				  BlockNumber *index_pages_removed,
+				  LVRelStats *vacrelstats);
 static int lazy_vacuum_page(Relation onerel, BlockNumber blkno, Buffer buffer,
 				 int tupindex, LVRelStats *vacrelstats);
 static void lazy_truncate_heap(Relation onerel, LVRelStats *vacrelstats);
@@ -169,7 +168,7 @@ lazy_vacuum_rel(Relation onerel, VacuumStmt *vacstmt)
 	 */
 	possibly_freeable = vacrelstats->rel_pages - vacrelstats->nonempty_pages;
 	if (possibly_freeable >= REL_TRUNCATE_MINIMUM ||
-	 possibly_freeable >= vacrelstats->rel_pages / REL_TRUNCATE_FRACTION)
+		possibly_freeable >= vacrelstats->rel_pages / REL_TRUNCATE_FRACTION)
 		lazy_truncate_heap(onerel, vacrelstats);
 
 	/* Update shared free space map with final free space info */
@@ -180,6 +179,10 @@ lazy_vacuum_rel(Relation onerel, VacuumStmt *vacstmt)
 						vacrelstats->rel_pages,
 						vacrelstats->rel_tuples,
 						hasindex);
+
+	/* report results to the stats collector, too */
+	pgstat_report_vacuum(RelationGetRelid(onerel), onerel->rd_rel->relisshared,
+						 vacstmt->analyze, vacrelstats->rel_tuples);
 }
 
 
@@ -208,9 +211,9 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 	BlockNumber *index_pages_removed;
 	bool		did_vacuum_index = false;
 	int			i;
-	VacRUsage	ru0;
+	PGRUsage	ru0;
 
-	vac_init_rusage(&ru0);
+	pg_rusage_init(&ru0);
 
 	relname = RelationGetRelationName(onerel);
 	ereport(elevel,
@@ -226,7 +229,7 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 	 * track of the total number of rows and pages removed from each index.
 	 * index_tups_vacuumed[i] is the number removed so far from the i'th
 	 * index.  (For partial indexes this could well be different from
-	 * tups_vacuumed.)  Likewise for index_pages_removed[i].
+	 * tups_vacuumed.)	Likewise for index_pages_removed[i].
 	 */
 	index_tups_vacuumed = (double *) palloc0(nindexes * sizeof(double));
 	index_pages_removed = (BlockNumber *) palloc0(nindexes * sizeof(BlockNumber));
@@ -251,11 +254,10 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 		vacuum_delay_point();
 
 		/*
-		 * If we are close to overrunning the available space for
-		 * dead-tuple TIDs, pause and do a cycle of vacuuming before we
-		 * tackle this page.
+		 * If we are close to overrunning the available space for dead-tuple
+		 * TIDs, pause and do a cycle of vacuuming before we tackle this page.
 		 */
-		if ((vacrelstats->max_dead_tuples - vacrelstats->num_dead_tuples) < MAX_TUPLES_PER_PAGE &&
+		if ((vacrelstats->max_dead_tuples - vacrelstats->num_dead_tuples) < MaxHeapTuplesPerPage &&
 			vacrelstats->num_dead_tuples > 0)
 		{
 			/* Remove index entries */
@@ -281,35 +283,35 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 		if (PageIsNew(page))
 		{
 			/*
-			 * An all-zeroes page could be left over if a backend extends
-			 * the relation but crashes before initializing the page.
-			 * Reclaim such pages for use.
+			 * An all-zeroes page could be left over if a backend extends the
+			 * relation but crashes before initializing the page. Reclaim such
+			 * pages for use.
 			 *
-			 * We have to be careful here because we could be looking at
-			 * a page that someone has just added to the relation and not
-			 * yet been able to initialize (see RelationGetBufferForTuple).
-			 * To interlock against that, release the buffer read lock
-			 * (which we must do anyway) and grab the relation extension
-			 * lock before re-locking in exclusive mode.  If the page is
-			 * still uninitialized by then, it must be left over from a
-			 * crashed backend, and we can initialize it.
+			 * We have to be careful here because we could be looking at a
+			 * page that someone has just added to the relation and not yet
+			 * been able to initialize (see RelationGetBufferForTuple). To
+			 * interlock against that, release the buffer read lock (which we
+			 * must do anyway) and grab the relation extension lock before
+			 * re-locking in exclusive mode.  If the page is still
+			 * uninitialized by then, it must be left over from a crashed
+			 * backend, and we can initialize it.
 			 *
-			 * We don't really need the relation lock when this is a new
-			 * or temp relation, but it's probably not worth the code space
-			 * to check that, since this surely isn't a critical path.
+			 * We don't really need the relation lock when this is a new or
+			 * temp relation, but it's probably not worth the code space to
+			 * check that, since this surely isn't a critical path.
 			 *
-			 * Note: the comparable code in vacuum.c need not do all this
-			 * because it's got exclusive lock on the whole relation.
+			 * Note: the comparable code in vacuum.c need not worry because
+			 * it's got exclusive lock on the whole relation.
 			 */
 			LockBuffer(buf, BUFFER_LOCK_UNLOCK);
-			LockPage(onerel, 0, ExclusiveLock);
-			UnlockPage(onerel, 0, ExclusiveLock);
+			LockRelationForExtension(onerel, ExclusiveLock);
+			UnlockRelationForExtension(onerel, ExclusiveLock);
 			LockBuffer(buf, BUFFER_LOCK_EXCLUSIVE);
 			if (PageIsNew(page))
 			{
 				ereport(WARNING,
-						(errmsg("relation \"%s\" page %u is uninitialized --- fixing",
-								relname, blkno)));
+				(errmsg("relation \"%s\" page %u is uninitialized --- fixing",
+						relname, blkno)));
 				PageInit(page, BufferGetPageSize(buf), 0);
 				empty_pages++;
 				lazy_record_free_space(vacrelstats, blkno,
@@ -363,15 +365,15 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 				case HEAPTUPLE_LIVE:
 
 					/*
-					 * Tuple is good.  Consider whether to replace its
-					 * xmin value with FrozenTransactionId.
+					 * Tuple is good.  Consider whether to replace its xmin
+					 * value with FrozenTransactionId.
 					 *
 					 * NB: Since we hold only a shared buffer lock here, we
-					 * are assuming that TransactionId read/write is
-					 * atomic.	This is not the only place that makes such
-					 * an assumption.  It'd be possible to avoid the
-					 * assumption by momentarily acquiring exclusive lock,
-					 * but for the moment I see no need to.
+					 * are assuming that TransactionId read/write is atomic.
+					 * This is not the only place that makes such an
+					 * assumption. It'd be possible to avoid the assumption by
+					 * momentarily acquiring exclusive lock, but for the
+					 * moment I see no need to.
 					 */
 					if (TransactionIdIsNormal(HeapTupleHeaderGetXmin(tuple.t_data)) &&
 						TransactionIdPrecedes(HeapTupleHeaderGetXmin(tuple.t_data),
@@ -394,8 +396,8 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 				case HEAPTUPLE_RECENTLY_DEAD:
 
 					/*
-					 * If tuple is recently deleted then we must not
-					 * remove it from relation.
+					 * If tuple is recently deleted then we must not remove it
+					 * from relation.
 					 */
 					nkeep += 1;
 					break;
@@ -424,9 +426,9 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 
 		/*
 		 * If we remembered any tuples for deletion, then the page will be
-		 * visited again by lazy_vacuum_heap, which will compute and
-		 * record its post-compaction free space.  If not, then we're done
-		 * with this page, so remember its free space as-is.
+		 * visited again by lazy_vacuum_heap, which will compute and record
+		 * its post-compaction free space.	If not, then we're done with this
+		 * page, so remember its free space as-is.
 		 */
 		if (vacrelstats->num_dead_tuples == prev_dead_count)
 		{
@@ -477,11 +479,11 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 			 errdetail("%.0f dead row versions cannot be removed yet.\n"
 					   "There were %.0f unused item pointers.\n"
 					   "%u pages are entirely empty.\n"
-					   "%s",
+					   "%s.",
 					   nkeep,
 					   nunused,
 					   empty_pages,
-					   vac_show_rusage(&ru0))));
+					   pg_rusage_show(&ru0))));
 }
 
 
@@ -501,9 +503,9 @@ lazy_vacuum_heap(Relation onerel, LVRelStats *vacrelstats)
 {
 	int			tupindex;
 	int			npages;
-	VacRUsage	ru0;
+	PGRUsage	ru0;
 
-	vac_init_rusage(&ru0);
+	pg_rusage_init(&ru0);
 	npages = 0;
 
 	tupindex = 0;
@@ -532,8 +534,8 @@ lazy_vacuum_heap(Relation onerel, LVRelStats *vacrelstats)
 			(errmsg("\"%s\": removed %d row versions in %d pages",
 					RelationGetRelationName(onerel),
 					tupindex, npages),
-			 errdetail("%s",
-					   vac_show_rusage(&ru0))));
+			 errdetail("%s.",
+					   pg_rusage_show(&ru0))));
 }
 
 /*
@@ -550,7 +552,7 @@ static int
 lazy_vacuum_page(Relation onerel, BlockNumber blkno, Buffer buffer,
 				 int tupindex, LVRelStats *vacrelstats)
 {
-	OffsetNumber unused[BLCKSZ / sizeof(OffsetNumber)];
+	OffsetNumber unused[MaxOffsetNumber];
 	int			uncnt;
 	Page		page = BufferGetPage(buffer);
 	ItemId		itemid;
@@ -601,13 +603,13 @@ lazy_scan_index(Relation indrel, LVRelStats *vacrelstats)
 {
 	IndexBulkDeleteResult *stats;
 	IndexVacuumCleanupInfo vcinfo;
-	VacRUsage	ru0;
+	PGRUsage	ru0;
 
-	vac_init_rusage(&ru0);
+	pg_rusage_init(&ru0);
 
 	/*
-	 * Acquire appropriate type of lock on index: must be exclusive if
-	 * index AM isn't concurrent-safe.
+	 * Acquire appropriate type of lock on index: must be exclusive if index
+	 * AM isn't concurrent-safe.
 	 */
 	if (indrel->rd_am->amconcurrent)
 		LockRelation(indrel, RowExclusiveLock);
@@ -616,9 +618,9 @@ lazy_scan_index(Relation indrel, LVRelStats *vacrelstats)
 
 	/*
 	 * Even though we're not planning to delete anything, we use the
-	 * ambulkdelete call, because (a) the scan happens within the index AM
-	 * for more speed, and (b) it may want to pass private statistics to
-	 * the amvacuumcleanup call.
+	 * ambulkdelete call, because (a) the scan happens within the index AM for
+	 * more speed, and (b) it may want to pass private statistics to the
+	 * amvacuumcleanup call.
 	 */
 	stats = index_bulk_delete(indrel, dummy_tid_reaped, NULL);
 
@@ -646,14 +648,14 @@ lazy_scan_index(Relation indrel, LVRelStats *vacrelstats)
 						false);
 
 	ereport(elevel,
-	   (errmsg("index \"%s\" now contains %.0f row versions in %u pages",
-			   RelationGetRelationName(indrel),
-			   stats->num_index_tuples,
-			   stats->num_pages),
-		errdetail("%u index pages have been deleted, %u are currently reusable.\n"
-				  "%s",
-				  stats->pages_deleted, stats->pages_free,
-				  vac_show_rusage(&ru0))));
+			(errmsg("index \"%s\" now contains %.0f row versions in %u pages",
+					RelationGetRelationName(indrel),
+					stats->num_index_tuples,
+					stats->num_pages),
+	errdetail("%u index pages have been deleted, %u are currently reusable.\n"
+			  "%s.",
+			  stats->pages_deleted, stats->pages_free,
+			  pg_rusage_show(&ru0))));
 
 	pfree(stats);
 }
@@ -678,13 +680,13 @@ lazy_vacuum_index(Relation indrel,
 {
 	IndexBulkDeleteResult *stats;
 	IndexVacuumCleanupInfo vcinfo;
-	VacRUsage	ru0;
+	PGRUsage	ru0;
 
-	vac_init_rusage(&ru0);
+	pg_rusage_init(&ru0);
 
 	/*
-	 * Acquire appropriate type of lock on index: must be exclusive if
-	 * index AM isn't concurrent-safe.
+	 * Acquire appropriate type of lock on index: must be exclusive if index
+	 * AM isn't concurrent-safe.
 	 */
 	if (indrel->rd_am->amconcurrent)
 		LockRelation(indrel, RowExclusiveLock);
@@ -722,16 +724,16 @@ lazy_vacuum_index(Relation indrel,
 						false);
 
 	ereport(elevel,
-	   (errmsg("index \"%s\" now contains %.0f row versions in %u pages",
-			   RelationGetRelationName(indrel),
-			   stats->num_index_tuples,
-			   stats->num_pages),
-		errdetail("%.0f index row versions were removed.\n"
-		 "%u index pages have been deleted, %u are currently reusable.\n"
-				  "%s",
-				  stats->tuples_removed,
-				  stats->pages_deleted, stats->pages_free,
-				  vac_show_rusage(&ru0))));
+			(errmsg("index \"%s\" now contains %.0f row versions in %u pages",
+					RelationGetRelationName(indrel),
+					stats->num_index_tuples,
+					stats->num_pages),
+			 errdetail("%.0f index row versions were removed.\n"
+			 "%u index pages have been deleted, %u are currently reusable.\n"
+					   "%s.",
+					   stats->tuples_removed,
+					   stats->pages_deleted, stats->pages_free,
+					   pg_rusage_show(&ru0))));
 
 	pfree(stats);
 }
@@ -748,24 +750,23 @@ lazy_truncate_heap(Relation onerel, LVRelStats *vacrelstats)
 	int			n;
 	int			i,
 				j;
-	VacRUsage	ru0;
+	PGRUsage	ru0;
 
-	vac_init_rusage(&ru0);
+	pg_rusage_init(&ru0);
 
 	/*
-	 * We need full exclusive lock on the relation in order to do
-	 * truncation. If we can't get it, give up rather than waiting --- we
-	 * don't want to block other backends, and we don't want to deadlock
-	 * (which is quite possible considering we already hold a lower-grade
-	 * lock).
+	 * We need full exclusive lock on the relation in order to do truncation.
+	 * If we can't get it, give up rather than waiting --- we don't want to
+	 * block other backends, and we don't want to deadlock (which is quite
+	 * possible considering we already hold a lower-grade lock).
 	 */
 	if (!ConditionalLockRelation(onerel, AccessExclusiveLock))
 		return;
 
 	/*
 	 * Now that we have exclusive lock, look to see if the rel has grown
-	 * whilst we were vacuuming with non-exclusive lock.  If so, give up;
-	 * the newly added pages presumably contain non-deletable tuples.
+	 * whilst we were vacuuming with non-exclusive lock.  If so, give up; the
+	 * newly added pages presumably contain non-deletable tuples.
 	 */
 	new_rel_pages = RelationGetNumberOfBlocks(onerel);
 	if (new_rel_pages != old_rel_pages)
@@ -778,9 +779,9 @@ lazy_truncate_heap(Relation onerel, LVRelStats *vacrelstats)
 
 	/*
 	 * Scan backwards from the end to verify that the end pages actually
-	 * contain nothing we need to keep.  This is *necessary*, not
-	 * optional, because other backends could have added tuples to these
-	 * pages whilst we were vacuuming.
+	 * contain no tuples.  This is *necessary*, not optional, because other
+	 * backends could have added tuples to these pages whilst we were
+	 * vacuuming.
 	 */
 	new_rel_pages = count_nondeletable_pages(onerel, vacrelstats);
 
@@ -793,16 +794,6 @@ lazy_truncate_heap(Relation onerel, LVRelStats *vacrelstats)
 
 	/*
 	 * Okay to truncate.
-	 *
-	 * First, flush any shared buffers for the blocks we intend to delete.
-	 * FlushRelationBuffers is a bit more than we need for this, since it
-	 * will also write out dirty buffers for blocks we aren't deleting,
-	 * but it's the closest thing in bufmgr's API.
-	 */
-	FlushRelationBuffers(onerel, new_rel_pages);
-
-	/*
-	 * Do the physical truncation.
 	 */
 	RelationTruncate(onerel, new_rel_pages);
 
@@ -837,8 +828,8 @@ lazy_truncate_heap(Relation onerel, LVRelStats *vacrelstats)
 			(errmsg("\"%s\": truncated %u to %u pages",
 					RelationGetRelationName(onerel),
 					old_rel_pages, new_rel_pages),
-			 errdetail("%s",
-					   vac_show_rusage(&ru0))));
+			 errdetail("%s.",
+					   pg_rusage_show(&ru0))));
 }
 
 /*
@@ -920,8 +911,8 @@ count_nondeletable_pages(Relation onerel, LVRelStats *vacrelstats)
 
 	/*
 	 * If we fall out of the loop, all the previously-thought-to-be-empty
-	 * pages really are; we need not bother to look at the last
-	 * known-nonempty page.
+	 * pages still are; we need not bother to look at the last known-nonempty
+	 * page.
 	 */
 	return vacrelstats->nonempty_pages;
 }
@@ -934,20 +925,22 @@ count_nondeletable_pages(Relation onerel, LVRelStats *vacrelstats)
 static void
 lazy_space_alloc(LVRelStats *vacrelstats, BlockNumber relblocks)
 {
-	int			maxtuples;
+	long		maxtuples;
 	int			maxpages;
 
-	maxtuples = (int) ((maintenance_work_mem * 1024L) / sizeof(ItemPointerData));
+	maxtuples = (maintenance_work_mem * 1024L) / sizeof(ItemPointerData);
+	maxtuples = Min(maxtuples, INT_MAX);
+	maxtuples = Min(maxtuples, MaxAllocSize / sizeof(ItemPointerData));
 	/* stay sane if small maintenance_work_mem */
-	if (maxtuples < MAX_TUPLES_PER_PAGE)
-		maxtuples = MAX_TUPLES_PER_PAGE;
+	maxtuples = Max(maxtuples, MaxHeapTuplesPerPage);
 
 	vacrelstats->num_dead_tuples = 0;
-	vacrelstats->max_dead_tuples = maxtuples;
+	vacrelstats->max_dead_tuples = (int) maxtuples;
 	vacrelstats->dead_tuples = (ItemPointer)
 		palloc(maxtuples * sizeof(ItemPointerData));
 
 	maxpages = MaxFSMPages;
+	maxpages = Min(maxpages, MaxAllocSize / sizeof(PageFreeSpaceInfo));
 	/* No need to allocate more pages than the relation has blocks */
 	if (relblocks < (BlockNumber) maxpages)
 		maxpages = (int) relblocks;
@@ -992,18 +985,16 @@ lazy_record_free_space(LVRelStats *vacrelstats,
 	/*
 	 * A page with less than stats->threshold free space will be forgotten
 	 * immediately, and never passed to the free space map.  Removing the
-	 * uselessly small entries early saves cycles, and in particular
-	 * reduces the amount of time we spend holding the FSM lock when we
-	 * finally call RecordRelationFreeSpace.  Since the FSM will probably
-	 * drop pages with little free space anyway, there's no point in
-	 * making this really small.
+	 * uselessly small entries early saves cycles, and in particular reduces
+	 * the amount of time we spend holding the FSM lock when we finally call
+	 * RecordRelationFreeSpace.  Since the FSM will probably drop pages with
+	 * little free space anyway, there's no point in making this really small.
 	 *
-	 * XXX Is it worth trying to measure average tuple size, and using that
-	 * to adjust the threshold?  Would be worthwhile if FSM has no stats
-	 * yet for this relation.  But changing the threshold as we scan the
-	 * rel might lead to bizarre behavior, too.  Also, it's probably
-	 * better if vacuum.c has the same thresholding behavior as we do
-	 * here.
+	 * XXX Is it worth trying to measure average tuple size, and using that to
+	 * adjust the threshold?  Would be worthwhile if FSM has no stats yet for
+	 * this relation.  But changing the threshold as we scan the rel might
+	 * lead to bizarre behavior, too.  Also, it's probably better if vacuum.c
+	 * has the same thresholding behavior as we do here.
 	 */
 	if (avail < vacrelstats->threshold)
 		return;
@@ -1037,8 +1028,8 @@ lazy_record_free_space(LVRelStats *vacrelstats,
 	{
 		/*
 		 * Scan backwards through the array, "sift-up" each value into its
-		 * correct position.  We can start the scan at n/2-1 since each
-		 * entry above that position has no children to worry about.
+		 * correct position.  We can start the scan at n/2-1 since each entry
+		 * above that position has no children to worry about.
 		 */
 		int			l = n / 2;
 
@@ -1074,9 +1065,9 @@ lazy_record_free_space(LVRelStats *vacrelstats,
 	{
 		/*
 		 * Notionally, we replace the zero'th entry with the new data, and
-		 * then sift-up to maintain the heap property.	Physically, the
-		 * new data doesn't get stored into the arrays until we find the
-		 * right location for it.
+		 * then sift-up to maintain the heap property.	Physically, the new
+		 * data doesn't get stored into the arrays until we find the right
+		 * location for it.
 		 */
 		int			i = 0;		/* i is where the "hole" is */
 
